@@ -1,5 +1,3 @@
-(* $Id: bi_io.ml 58 2010-11-15 00:23:24Z mjambon $ *)
-
 open Printf
 
 open Bi_outbuf
@@ -12,6 +10,7 @@ let int8_tag = 1
 let int16_tag = 2
 let int32_tag = 3
 let int64_tag = 4
+let float32_tag = 11
 let float64_tag = 12
 let uvint_tag = 16
 let svint_tag = 17
@@ -23,7 +22,7 @@ let num_variant_tag = 22
 let variant_tag = 23
 let unit_tag = 24
 let table_tag = 25
-let shared_tag = 26 
+let shared_tag = 26
 
 type hash = int
 
@@ -37,6 +36,7 @@ type tree =
     | `Int16 of int
     | `Int32 of Int32.t
     | `Int64 of Int64.t
+    | `Float32 of float
     | `Float64 of float
     | `Uvint of int
     | `Svint of int
@@ -46,10 +46,10 @@ type tree =
     | `Record of (string option * hash * tree) array
     | `Num_variant of (int * tree option)
     | `Variant of (string option * hash * tree option)
-    | `Table of 
+    | `Table of
 	((string option * hash * node_tag) array * tree array array) option
     | `Shared of tree ]
-    
+
 (* extend sign bit *)
 let make_signed x =
   if x > 0x3FFFFFFF then x - (1 lsl 31) else x
@@ -111,7 +111,7 @@ let read_hashtag ib cont =
   let x3 = (Char.code s.[i+2]) lsl 8 in
   let x4 = Char.code s.[i+3] in
   let h = make_signed (x1 lor x2 lor x3 lor x4) in
-  
+
   cont ib h has_arg
 
 
@@ -126,7 +126,7 @@ let read_field_hashtag ib =
   let x3 = (Char.code (String.unsafe_get s (i+2))) lsl 8 in
   let x4 = Char.code (String.unsafe_get s (i+3)) in
   make_signed (x1 lor x2 lor x3 lor x4)
-  
+
 
 type int7 = int
 
@@ -150,7 +150,7 @@ let make_unhash l =
   List.iter (
     fun s ->
       let h = hash_name s in
-      try 
+      try
 	match Hashtbl.find tbl h with
 	    Some s' ->
 	      if s <> s' then
@@ -195,18 +195,22 @@ let write_untagged_int32 ob x =
   let low = Int32.to_int x in
   Bi_outbuf.add_char ob (Char.chr ((low lsr 8) land 0xff));
   Bi_outbuf.add_char ob (Char.chr (low land 0xff))
-    
-let float_endianness =
+
+let write_untagged_float32 ob x =
+  write_untagged_int32 ob (Int32.bits_of_float x)
+
+let float_endianness = lazy (
   match String.unsafe_get (Obj.magic 1.0) 0 with
       '\x3f' -> `Big
     | '\x00' -> `Little
     | _ -> assert false
+)
 
 let read_untagged_float64 ib =
   let i = Bi_inbuf.read ib 8 in
   let s = ib.i_s in
   let x = Obj.new_block Obj.double_tag 8 in
-  (match float_endianness with
+  (match Lazy.force float_endianness with
        `Little ->
 	 for j = 0 to 7 do
 	   String.unsafe_set (Obj.obj x) (7-j) (String.unsafe_get s (i+j))
@@ -221,7 +225,7 @@ let read_untagged_float64 ib =
 let write_untagged_float64 ob x =
   let i = Bi_outbuf.alloc ob 8 in
   let s = ob.o_s in
-  (match float_endianness with
+  (match Lazy.force float_endianness with
        `Little ->
 	 for j = 0 to 7 do
 	   String.unsafe_set s (i+j) (String.unsafe_get (Obj.magic x) (7-j))
@@ -252,7 +256,7 @@ let write_untagged_int64 ob x =
   write_untagged_float64 ob (Int64.float_of_bits x)
 
 
-let () =
+let safety_test () =
   let s = "\x3f\xf0\x06\x05\x04\x03\x02\x01" in
   let x = 1.00146962706651288 in
   let y = read_untagged_float64 (Bi_inbuf.from_string s) in
@@ -300,6 +304,10 @@ let write_int64 ob x =
   write_tag ob int64_tag;
   write_untagged_int64 ob x
 
+let write_float32 ob x =
+  write_tag ob float32_tag;
+  write_untagged_float32 ob x
+
 let write_float64 ob x =
   write_tag ob float64_tag;
   write_untagged_float64 ob x
@@ -332,7 +340,7 @@ let rec write_t ob tagged (x : tree) =
 	write_untagged_bool ob x
 
     | `Int8 x ->
-	if tagged then 
+	if tagged then
 	  write_tag ob int8_tag;
 	write_untagged_char ob x
 
@@ -351,6 +359,11 @@ let rec write_t ob tagged (x : tree) =
 	  write_tag ob int64_tag;
 	write_untagged_int64 ob x
 
+    | `Float32 x ->
+        if tagged then
+	  write_tag ob float32_tag;
+	write_untagged_float32 ob x
+
     | `Float64 x ->
 	if tagged then
 	  write_tag ob float64_tag;
@@ -471,6 +484,7 @@ let tag_of_tree (x : tree) =
     | `Int16 _ -> int16_tag
     | `Int32 _ -> int32_tag
     | `Int64 _ -> int64_tag
+    | `Float32 _ -> float32_tag
     | `Float64 _ -> float64_tag
     | `Uvint _ -> uvint_tag
     | `Svint _ -> svint_tag
@@ -484,9 +498,6 @@ let tag_of_tree (x : tree) =
     | `Shared _ -> shared_tag
 
 
-let tag_error () =
-  Bi_util.error "Corrupted data (tag)"
-
 let read_tag ib =
   Char.code (Bi_inbuf.read_char ib)
 
@@ -521,6 +532,9 @@ let read_untagged_int32 ib =
     Int32.of_int (((Char.code s.[i+2]) lsl 8) lor (Char.code s.[i+3])) in
   Int32.logor (Int32.shift_left x1 16) x2
 
+let read_untagged_float32 ib =
+  Int32.float_of_bits (read_untagged_int32 ib)
+
 (*
 let read_untagged_int64 ib =
   let i = Bi_inbuf.read ib 8 in
@@ -576,7 +590,10 @@ let read_int32 ib = `Int32 (read_untagged_int32 ib)
 
 let read_int64 ib = `Int64 (read_untagged_int64 ib)
 
-let read_float ib =
+let read_float32 ib =
+  `Float32 (read_untagged_float32 ib)
+
+let read_float64 ib =
   `Float64 (read_untagged_float64 ib)
 
 let read_uvint ib = `Uvint (read_untagged_uvint ib)
@@ -595,21 +612,21 @@ let read_tree ?(unhash = make_unhash []) ib : tree =
       let tag = read_tag ib in
       let read = reader_of_tag tag in
       `Array (Some (tag, Array.init len (fun _ -> read ib)))
-      
+
   and read_tuple ib =
     let len = Bi_vint.read_uvint ib in
     `Tuple (Array.init len (fun _ -> read_tree ib))
-      
+
   and read_field ib =
     let h = read_field_hashtag ib in
     let name = unhash h in
     let x = read_tree ib in
     (name, h, x)
-      
+
   and read_record ib =
     let len = Bi_vint.read_uvint ib in
     `Record (Array.init len (fun _ -> read_field ib))
-    
+
   and read_num_variant_cont ib i has_arg =
     let x =
       if has_arg then
@@ -618,10 +635,10 @@ let read_tree ?(unhash = make_unhash []) ib : tree =
 	None
     in
     `Num_variant (i, x)
-  
+
   and read_num_variant ib =
     read_numtag ib read_num_variant_cont
-      
+
   and read_variant_cont ib h has_arg =
     let name = unhash h in
     let x =
@@ -631,17 +648,17 @@ let read_tree ?(unhash = make_unhash []) ib : tree =
 	None
     in
     `Variant (name, h, x)
-  
+
   and read_variant ib =
     read_hashtag ib read_variant_cont
-      
+
   and read_table ib =
     let row_num = Bi_vint.read_uvint ib in
     if row_num = 0 then
       `Table None
     else
       let col_num = Bi_vint.read_uvint ib in
-      let fields = 
+      let fields =
 	Array.init col_num (
 	  fun _ ->
 	    let h = read_field_hashtag ib in
@@ -650,7 +667,7 @@ let read_tree ?(unhash = make_unhash []) ib : tree =
 	    (name, h, tag)
 	)
       in
-      let readers = 
+      let readers =
 	Array.map (fun (name, h, tag) -> reader_of_tag tag) fields in
       let a =
 	Array.init row_num
@@ -658,7 +675,7 @@ let read_tree ?(unhash = make_unhash []) ib : tree =
 	     Array.init col_num (fun j -> readers.(j) ib))
       in
       `Table (Some (fields, a))
-	
+
   and read_shared ib =
     let pos = ib.i_offs + ib.i_pos in
     let offset = Bi_vint.read_uvint ib in
@@ -679,7 +696,8 @@ let read_tree ?(unhash = make_unhash []) ib : tree =
     | 2 (* int16 *) -> read_int16
     | 3 (* int32 *) -> read_int32
     | 4 (* int64 *) -> read_int64
-    | 12 (* float *) -> read_float
+    | 11 (* float32 *) -> read_float32
+    | 12 (* float64 *) -> read_float64
     | 16 (* uvint *) -> read_uvint
     | 17 (* svint *) -> read_svint
     | 18 (* string *) -> read_string
@@ -692,10 +710,10 @@ let read_tree ?(unhash = make_unhash []) ib : tree =
     | 25 (* table *) -> read_table
     | 26 (* shared *) -> read_shared
     | _ -> Bi_util.error "Corrupted data (invalid tag)"
-	
+
   and read_tree ib : tree =
     reader_of_tag (read_tag ib) ib
-      
+
   in
   read_tree ib
 
@@ -710,7 +728,8 @@ let skip_int8 ib = skip_bytes ib 1
 let skip_int16 ib = skip_bytes ib 2
 let skip_int32 ib = skip_bytes ib 4
 let skip_int64 ib = skip_bytes ib 8
-let skip_float ib = skip_bytes ib 8
+let skip_float32 ib = skip_bytes ib 4
+let skip_float64 ib = skip_bytes ib 8
 let skip_uvint ib = ignore (read_untagged_uvint ib)
 let skip_svint ib = ignore (read_untagged_svint ib)
 
@@ -727,44 +746,44 @@ let rec skip_array ib =
     for i = 1 to len do
       read ib
     done
-      
+
 and skip_tuple ib =
   let len = Bi_vint.read_uvint ib in
   for i = 1 to len do
     skip ib
   done
-    
+
 and skip_field ib =
   ignore (read_field_hashtag ib);
   skip ib
-    
+
 and skip_record ib =
   let len = Bi_vint.read_uvint ib in
   for i = 1 to len do
     skip_field ib
   done
-    
+
 and skip_num_variant_cont ib i has_arg =
   if has_arg then
     skip ib
-      
+
 and skip_num_variant ib =
   read_numtag ib skip_num_variant_cont
-    
+
 and skip_variant_cont ib h has_arg =
   if has_arg then
     skip ib
-      
+
 and skip_variant ib =
   read_hashtag ib skip_variant_cont
-    
+
 and skip_table ib =
   let row_num = Bi_vint.read_uvint ib in
   if row_num = 0 then
     ()
   else
     let col_num = Bi_vint.read_uvint ib in
-    let readers = 
+    let readers =
       Array.init col_num (
 	fun _ ->
 	  ignore (read_field_hashtag ib);
@@ -776,14 +795,15 @@ and skip_table ib =
 	readers.(j) ib
       done
     done
-      
+
 and skipper_of_tag = function
     0 (* bool *) -> skip_bool
   | 1 (* int8 *) -> skip_int8
   | 2 (* int16 *) -> skip_int16
   | 3 (* int32 *) -> skip_int32
   | 4 (* int64 *) -> skip_int64
-  | 12 (* float *) -> skip_float
+  | 11 (* float32 *) -> skip_float32
+  | 12 (* float64 *) -> skip_float64
   | 16 (* uvint *) -> skip_uvint
   | 17 (* svint *) -> skip_svint
   | 18 (* string *) -> skip_string
@@ -795,10 +815,10 @@ and skipper_of_tag = function
   | 24 (* unit *) -> skip_unit
   | 25 (* table *) -> skip_table
   | _ -> Bi_util.error "Corrupted data (invalid tag)"
-	
+
 and skip ib : unit =
   skipper_of_tag (read_tag ib) ib
-    
+
 
 (* Equivalent of Array.map that guarantees a left-to-right order *)
 let array_map f a =
@@ -836,6 +856,7 @@ struct
       | `Int16 x -> Atom (sprintf "0x%04x" x, atom)
       | `Int32 x -> Atom (sprintf "0x%08lx" x, atom)
       | `Int64 x -> Atom (sprintf "0x%016Lx" x, atom)
+      | `Float32 x -> Atom (string_of_float x, atom)
       | `Float64 x -> Atom (string_of_float x, atom)
       | `Uvint x -> Atom (string_of_int x, atom)
       | `Svint x -> Atom (string_of_int x, atom)
@@ -878,7 +899,7 @@ struct
 		  fun a ->
 		    `Record (
 		      Array.mapi (
-			fun i x -> 
+			fun i x ->
 			  let s, h, _ = header.(i) in
 			  (s, h, x)
 		      ) a
@@ -887,7 +908,7 @@ struct
 	      )
 	    ) in
 	  format shared record_array
-	    
+
       | `Shared x ->
           let tbl, p = shared in
           incr p;
@@ -910,14 +931,20 @@ end
 
 let init () = (Bi_share.Wr.create 512, ref 0)
 
+let view_of_tree t =
+  Easy_format.Pretty.to_string (Pp.format (init ()) t)
+
+let print_view_of_tree t =
+  Easy_format.Pretty.to_stdout (Pp.format (init ()) t)
+
+let output_view_of_tree oc t =
+  Easy_format.Pretty.to_channel oc (Pp.format (init ()) t)
+
 let view ?unhash s =
-  Easy_format.Pretty.to_string
-    (Pp.format (init ()) (tree_of_string ?unhash s))
+  view_of_tree (tree_of_string ?unhash s)
 
 let print_view ?unhash s =
-  Easy_format.Pretty.to_stdout
-    (Pp.format (init ()) (tree_of_string ?unhash s))
+  print_view_of_tree (tree_of_string ?unhash s)
 
 let output_view ?unhash oc s =
-  Easy_format.Pretty.to_channel oc 
-    (Pp.format (init ()) (tree_of_string ?unhash s))
+  output_view_of_tree oc (tree_of_string ?unhash s)
